package chord

import (
	"errors"
	"net/netip"
	"sync"
)

// Node represents a node's topological relations within a Chord network.
// It is safe to use the methods of Node concurrently.
type Node struct {
	// mu guards the mutable fields of the node.
	mu sync.Mutex
	// self is the node's own address.
	// It is fully immutable; the mutex does not protect it.
	self Peer
	// pred is the node's predecessor.
	// If not valid, the predecessor is not currently known.
	pred Peer
	// succ is the node's immediate successors for replication. Its capacity
	// is the replication factor for the network.
	// There is always at least one successor. If the number of nodes in the
	// network is not larger than the replication factor, then this node will
	// appear in its own successor list.
	succ []Peer
	// fingers is the node's finger table.
	fingers []Peer

	// data is the data the node owns.
	data map[ID]string
}

func (n *Node) Self() Peer {
	return n.self
}

func (n *Node) Successor() Peer {
	n.mu.Lock()
	defer n.mu.Unlock()
	return n.succ[0]
}

// Neighbors returns the node's predecessor and appends its successor list to s.
func (n *Node) Neighbors(s []Peer) (Peer, []Peer) {
	n.mu.Lock()
	defer n.mu.Unlock()
	return n.pred, append(s, n.succ...)
}

func (n *Node) localLocked(id ID) bool {
	return contains(n.self.id, n.succ[0].id, id)
}

// IsLocal reports whether this node owns the given key.
func (n *Node) IsLocal(id ID) bool {
	n.mu.Lock()
	defer n.mu.Unlock()
	return n.localLocked(id)
}

// Closest finds the locally known peer which is the closest predecessor of key.
func (n *Node) Closest(id ID) Peer {
	self := n.self.id
	n.mu.Lock()
	defer n.mu.Unlock()
	l := n.fingers
	for i := len(l) - 1; i >= 0; i-- {
		f := l[i]
		if contains(self, id, f.id) {
			// contains reports true if id == f.id, which is
			// technically incorrect for this operation.
			// We could reasonably skip double-checking and rely on the hash,
			// but we can also not skip it and just be correct.
			if id == f.id {
				continue
			}
			return f
		}
	}
	// Also try successors.
	l = n.succ
	for i := len(l) - 1; i >= 0; i-- {
		f := l[i]
		if contains(self, id, f.id) {
			if id == f.id {
				continue
			}
			return f
		}
	}
	return n.self
}

// GetLocal obtains the value for a key if it is local to and owned by the node.
func (n *Node) GetLocal(k ID) (v string, found bool) {
	n.mu.Lock()
	defer n.mu.Unlock()
	if n.localLocked(k) {
		v, found = n.data[k]
	}
	return v, found
}

// SetLocal sets the value for a key.
// Returns false if the key is not owned by the node.
func (n *Node) SetLocal(k ID, v string) bool {
	n.mu.Lock()
	defer n.mu.Unlock()
	if n.localLocked(k) {
		n.data[k] = v
		return true
	}
	return false
}

// Peer is the ID and address of a node.
type Peer struct {
	id   ID
	addr netip.AddrPort
}

// Address constructs a peer from an address and port.
func Address(addr netip.AddrPort) Peer {
	return Peer{id: addrID(addr), addr: addr}
}

// IsValid reports whether the peer was created using [Address] with an address
// which was valid.
func (p Peer) IsValid() bool { return p.addr.IsValid() && p.addr.Port() != 0 }

// Values returns the peer's ID and address.
func (p Peer) Values() (ID, netip.AddrPort) {
	// We do this instead of exporting Peer's fields to ensure those fields
	// are never mutable. If p.IsValid() then p.id == addrID(p.addr) always.
	return p.id, p.addr
}

// Create creates a new Chord network using the given address as the initial node.
func Create(addr netip.AddrPort) (*Node, error) {
	if !addr.IsValid() {
		return nil, errors.New("chord: cannot create with invalid address")
	}
	self := Address(addr)
	n := &Node{
		self:    self,
		succ:    []Peer{self}[:1:1], // extra cautious about capacity
		fingers: make([]Peer, 0, len(ID{})*8),
	}
	return n, nil
}
